"""
SCGI XMLRPC Transport

https://github.com/JohnDoee/autotorrent/blob/develop/autotorrent/scgitransport.py

XMLRPC in Python only supports HTTP(S). This module extends the transport to
also support SCGI. SCGI is required by rTorrent if you want to communicate
directly with an instance.

Example:

    >>> from xmlrpc.client import ServerProxy
    >>> proxy = ServerProxy('http://127.0.0.1:8000/', transport=ScgiTransport())
    >>> proxy.system.listMethods()

License:
    Public Domain (no attribution needed).
    The license only applies to THIS file.
"""

import io
import socket
import xmlrpc.client


def encode_netstring(input):
    return str(len(input)).encode() + b':' + input + b','


def encode_header(key, value):
    return key + b'\x00' + value + b'\x00'


class ScgiTransport(xmlrpc.client.Transport):
    def __init__(self, *args, **kwargs):
        self.socket_path = kwargs.pop('socket_path', '')
        super().__init__(*args, **kwargs)

    def single_request(self, host, handler, request_body, verbose=False):
        self.verbose = verbose
        if self.socket_path:
            s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            s.connect(self.socket_path)
        else:
            host, port = host.split(':')
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((host, int(port)))

        request = encode_header(b'CONTENT_LENGTH', str(len(request_body)).encode())
        request += encode_header(b'SCGI', b'1')
        request += encode_header(b'REQUEST_METHOD', b'POST')
        request += encode_header(b'REQUEST_URI', handler.encode())

        request = encode_netstring(request)
        request += request_body

        s.send(request)

        response = b''
        while True:
            r = s.recv(1024)
            if not r:
                break
            response += r

        response_body = io.BytesIO(b'\r\n\r\n'.join(response.split(b'\r\n\r\n')[1:]))
        return self.parse_response(response_body)

    def __repr__(self):
        return f'{type(self).__name__}(socket_path={self.socket_path!r})'



"""
Client API for qBittorrent
"""

import os
import urllib

from ... import errors, utils
from .. import asynccontextmanager, http
from .base import ClientApiBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class QbittorrentClientApi(ClientApiBase):
    """
    qBittorrent API

    Reference: https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)
    """

    name = 'qbittorrent'
    label = 'qBittorrent'

    default_config = {
        'url': 'http://localhost:8080',
        'username': '',
        'password': '',
        'check_after_add': utils.configfiles.config_value(
            value=utils.types.Bool('no'),
            description=f'Whether to tell {label} to verify pieces after adding the torrent.',
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        url = urllib.parse.urlparse(self.config['url'])
        self._headers = {
            'Referer': f'{url.scheme}://{url.netloc}',
        }

    async def _request(self, path, data=None, files={}):
        url = '/'.join((
            self.config['url'].rstrip('/'),
            'api/v2',
            path.lstrip('/'),
        ))
        return await http.post(
            url=url,
            headers=self._headers,
            data=data,
            files=files
        )

    async def _login(self):
        credentials = {
            'username': self.config['username'],
            'password': self.config['password'],
        }
        response = await self._request(path='auth/login', data=credentials)
        if response != 'Ok.':
            raise errors.RequestError('Authentication failed')

    async def _logout(self):
        await self._request(path='auth/logout')

    @asynccontextmanager
    async def _session(self):
        await self._login()
        try:
            yield
        finally:
            await self._logout()

    async def add_torrent(self, torrent_path, download_path=None):
        files = {
            'torrents': {
                'file': torrent_path,
                'mimetype': 'application/x-bittorrent',
            },
        }

        data = {
            'skip_checking': 'false' if self.config['check_after_add'] else 'true',
        }
        if download_path:
            data['savepath'] = str(os.path.abspath(download_path))

        async with self._session():
            response = await self._request('torrents/add', files=files, data=data)
            if response == 'Ok.':
                return self._get_torrent_hash(torrent_path)
            else:
                raise errors.RequestError('Unknown error')

    def _get_torrent_hash(self, torrent_path):
        # TODO: Get the info hash from `response` when this is closed:
        #       https://github.com/qbittorrent/qBittorrent/issues/4879
        #       For older versions of qBittorrent, default to getting the hash
        #       from the torrent file.
        try:
            return self.read_torrent(torrent_path).infohash
        except errors.TorrentError as e:
            raise errors.RequestError(e)



"""
Create torrent file
"""

import collections
import datetime
import hashlib
import math
import os
import time
from os.path import exists as _path_exists

from .. import __project_name__, __version__, constants, errors, utils
from . import LazyModule, fs, types

torf = LazyModule(module='torf', namespace=globals())


def create(*, content_path, announce, source, torrent_path,
           init_callback, progress_callback,
           overwrite=False, exclude=(), reuse_torrent_path=None):
    """
    Generate and write torrent file

    :param str content_path: Path to the torrent's payload
    :param str announce: Announce URL
    :param str source: Value of the ``source`` field in the torrent. This makes
        the torrent unique for each tracker to avoid cross-seeding issues, so it
        is usually the tracker's abbreviated name.
    :param str torrent_path: Path of the generated torrent file
    :param str init_callback: Callable that is called once before torrent
        generation commences. It gets `content_path` as a tree where each node
        is a tuple in which the first item is the directory name and the second
        item is a sequence of `(file_name, file_size)` tuples.

        Example:

        .. code::

           ('Parent',
               ('Foo', (
                   ('Picture.jpg', 82489),
                   ('Music.mp3', 5315672),
                   ('More files', (
                       ('This.txt', 57734),
                       ('And that.txt', 184),
                       ('Also some of this.txt', 88433),
                   )),
               )),
               ('Bar', (
                   ('Yee.mp4', 288489392),
                   ('Yah.mkv', 3883247384),
               )),
           )
    :param str progress_callback: Callable that is called at regular intervals
        with a :class:`CreateTorrentStatus` object as a positional argument.
        Torrent creation is cancelled if `progress_callback` returns `True` or
        any other truthy value.
    :param bool overwrite: Whether to overwrite `torrent_path` if it exists
    :param exclude: Sequence of regular expressions that are matched against
        file system paths. Matching files are not included in the torrent.
    :param reuse_torrent_path: Path to existing torrent file to get hashed
        pieces and piece size from. This argument is ignored if the files in the
        provided torrent do not match the files we want or if it can't be read
        for any reason.

    :raise TorrentError: if anything goes wrong

    :return: `torrent_path`
    """
    if not announce:
        raise errors.TorrentError('Announce URL is empty')
    if not source:
        raise errors.TorrentError('Source is empty')

    if overwrite or not _path_exists(torrent_path):
        torrent = None

        # Try to read piece hashes from cache
        torrent = _read_cache_torrent(
            content_path=content_path,
            cache_torrent_path=reuse_torrent_path,
            exclude=exclude,
        )
        if torrent:
            # Update metadata in generic torrent
            torrent.trackers = (announce,)
            torrent.source = source
            if reuse_torrent_path:
                _store_cache_torrent(torrent)
        else:
            # Reading cached piece hashes failed for some reason
            torrent = _generate_torrent(
                content_path=content_path,
                announce=announce,
                source=source,
                exclude=exclude,
                progress_callback=progress_callback,
                init_callback=init_callback,
            )
            if torrent.is_ready:
                _store_cache_torrent(torrent)

        # Write torrent file
        try:
            torrent.write(torrent_path, overwrite=True)
        except torf.TorfError as e:
            raise errors.TorrentError(e)

    return torrent_path


def _generate_torrent(*, content_path, announce, source, exclude, init_callback, progress_callback):
    callback = _CreateTorrentCallbackWrapper(progress_callback)
    try:
        torrent = torf.Torrent(
            path=content_path,
            exclude_regexs=exclude,
            trackers=((announce,),),
            private=True,
            source=source,
            created_by=f'{__project_name__} {__version__}',
            creation_date=time.time(),
        )
        init_callback(_make_file_tree(torrent.filetree))
        torrent.generate(callback=callback, interval=1.0)
    except torf.TorfError as e:
        raise errors.TorrentError(e)
    else:
        return torrent


def _store_cache_torrent(torrent):
    cache_torrent = torf.Torrent(
        private=True,
        created_by=f'{__project_name__} {__version__}',
        creation_date=time.time(),
        comment='This torrent is used to cache previously hashed pieces.',
    )
    _copy_torrent_info(torrent, cache_torrent)
    cache_torrent_path = _get_cache_torrent_path(cache_torrent, create_directory=True)
    cache_torrent.write(cache_torrent_path, overwrite=True)


def _read_cache_torrent(content_path, exclude, cache_torrent_path=None):
    try:
        # Create the Torrent we're going to use
        torrent = torf.Torrent(
            path=content_path,
            exclude_regexs=exclude,
            private=True,
            created_by=f'{__project_name__} {__version__}',
            creation_date=time.time(),
        )
        # Create the Torrent to get hashed pieces from
        if not cache_torrent_path:
            cache_torrent_path = _get_cache_torrent_path(torrent, create_directory=False)
        cache_torrent = torf.Torrent.read(cache_torrent_path)
    except torf.TorfError:
        pass
    else:
        # Make sure the cached torrent metainfo matches what we expect
        id_wanted = _get_torrent_id(torrent)
        id_cached = _get_torrent_id(cache_torrent)
        if id_cached == id_wanted:
            _copy_torrent_info(cache_torrent, torrent)
            return torrent


def _get_cache_torrent_path(torrent, create_directory=True):
    directory = constants.GENERIC_TORRENTS_DIRPATH
    if create_directory:
        try:
            fs.mkdir(directory)
        except errors.ContentError as e:
            raise errors.TorrentError(f'{directory}: {e}')

    cache_id = _get_torrent_id(torrent)
    filename = f'{torrent.name}.{cache_id}.torrent'
    return os.path.join(directory, filename)


def _get_torrent_id(torrent):
    id_info = {
        'name': torrent.name,
        'files': tuple((str(f), f.size) for f in torrent.files),
    }
    return utils.semantic_hash(id_info)


def _copy_torrent_info(from_torrent, to_torrent):
    from_info, to_info = from_torrent.metainfo['info'], to_torrent.metainfo['info']
    to_info['pieces'] = from_info['pieces']
    to_info['piece length'] = from_info['piece length']
    to_info['name'] = from_info['name']
    if 'length' in from_info:
        to_info['length'] = from_info['length']
    else:
        to_info['files'] = from_info['files']


def _make_file_tree(tree):
    files = []
    for name,file in tree.items():
        if isinstance(file, collections.abc.Mapping):
            subtree = _make_file_tree(file)
            files.append((name, subtree))
        else:
            files.append((name, file.size))
    return tuple(files)


class _CreateTorrentCallbackWrapper:
    def __init__(self, callback):
        self._callback = callback
        self._progress_samples = []
        self._time_started = datetime.datetime.now()

    def __call__(self, torrent, filepath, pieces_done, pieces_total):
        time_now = datetime.datetime.now()
        percent_done = pieces_done / pieces_total * 100
        seconds_elapsed = time_now - self._time_started

        def progress_samples_trim_condition(samples):
            oldest_sample_age = time_now - samples[0][0]
            return oldest_sample_age > datetime.timedelta(seconds=10)

        self._maintain_samples(
            samples=self._progress_samples,
            new_sample=(time_now, pieces_done),
            trim_condition=progress_samples_trim_condition,
        )

        # Estimate how long torrent creation will take
        if len(self._progress_samples) >= 2:
            # Calculate the difference between each pair of samples
            diffs = [
                b[1] - a[1]
                for a, b in zip(self._progress_samples[:-1], self._progress_samples[1:])
            ]
            pieces_per_second = self._get_average(diffs, weight_factor=1.1)
            bytes_per_second = types.Bytes(pieces_per_second * torrent.piece_size)
            pieces_remaining = pieces_total - pieces_done
            bytes_remaining = pieces_remaining * torrent.piece_size
            seconds_remaining = datetime.timedelta(seconds=bytes_remaining / bytes_per_second)
        else:
            bytes_per_second = types.Bytes(0)
            seconds_remaining = datetime.timedelta(seconds=0)

        seconds_total = seconds_elapsed + seconds_remaining
        time_finished = self._time_started + seconds_total

        status = CreateTorrentStatus(
            bytes_per_second=bytes_per_second,
            filepath=filepath,
            percent_done=percent_done,
            piece_size=torrent.piece_size,
            pieces_done=pieces_done,
            pieces_total=pieces_total,
            seconds_elapsed=seconds_elapsed,
            seconds_remaining=seconds_remaining,
            seconds_total=seconds_total,
            time_finished=time_finished,
            time_started=self._time_started,
            total_size=torrent.size,
        )
        return self._callback(status)

    def _maintain_samples(self, new_sample, samples, trim_condition):
        samples.append(new_sample)
        while samples and trim_condition(samples):
            del samples[0]

    def _get_average(self, samples, weight_factor, get_value=lambda sample: sample):
        # Give recent samples more weight than older samples
        # https://en.wikipedia.org/wiki/Moving_average
        weights = []
        for _ in range(len(samples)):
            try:
                weight = weights[-1]
            except IndexError:
                weight = 1
            weights.append(weight * weight_factor)

        return sum(
            get_value(sample) * weight
            for sample, weight in zip(samples, weights)
        ) / sum(weights)


class CreateTorrentStatus(collections.namedtuple(
    typename='CreateTorrentStatus',
    field_names=(
        'bytes_per_second',
        'filepath',
        'percent_done',
        'piece_size',
        'pieces_done',
        'pieces_total',
        'seconds_elapsed',
        'seconds_remaining',
        'seconds_total',
        'time_started',
        'time_finished',
        'total_size',
    ),
)):
    """
    :func:`~.collections.namedtuple` with these attributes:

        - ``bytes_per_second`` (:class:`~.types.Bytes`)
        - ``filepath`` (:class:`str`)
        - ``percent_done`` (:class:`float`)
        - ``piece_size`` (:class:`~.types.Bytes`)
        - ``pieces_done`` (:class:`int`)
        - ``pieces_total`` (:class:`int`)
        - ``seconds_elapsed`` (:class:`float`)
        - ``seconds_remaining`` (:class:`float`)
        - ``seconds_total`` (:class:`float`)
        - ``time_started`` (:class:`float`)
        - ``time_finished`` (:class:`float`)
        - ``total_size`` (:class:`~.types.Bytes`)
    """


class TorrentFileStream:
    """
    Traverse files as they are described in a torrent

    :param torrent: :class:`~.torf.Torrent` object
    :param location: Path to location of files in `torrent`

    Instances should be used as a context manager to ensure all files are closed
    properly.

    Example:
    >>> torrent = torf.Torrent(...)
    >>> with TorrentFileStream(torrent, location='path/to/files') as tfs:
    >>>     # Get the 29th piece of the concatenated file stream
    >>>     piece = tfs.get_piece(29)
    """

    def __init__(self, torrent, location):
        self._torrent = torrent
        self._location = location
        self._open_files = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        """
        Close all opened files

        This is called automatically when the instance is used as a context
        manager.
        """
        for filepath, fh in tuple(self._open_files.items()):
            fh.close()
            del self._open_files[filepath]

    def get_file_position(self, file):
        """
        Return index of first byte of `file` in stream of concatenated files

        :param file: :class:`~torf.File` object
        """
        file_index = self._torrent.files.index(file)
        stream_pos = sum(f.size for f in self._torrent.files[:file_index])
        return stream_pos

    def get_file_at_position(self, byte_index):
        """
        Return file that belongs to the byte at `byte_index` in stream of
        concatenated files

        :param byte_index: Index of the byte; minimum is 0, maximum is the
            torrent's size minus 1
        """
        if byte_index >= 0:
            pos = 0
            for f in self._torrent.files:
                pos += f.size - 1
                if pos >= byte_index:
                    return f
                else:
                    pos += 1
        raise ValueError(f'byte_index is out of bounds (0 - {self._torrent.size}): {byte_index}')

    def get_piece_indexes_for_file(self, file):
        """
        Return indexes of pieces that contain at least one byte of `file`

        :param file: :class:`~torf.File` object
        """
        stream_pos = self.get_file_position(file)
        first_piece_index = math.floor(stream_pos / self._torrent.piece_size)
        last_piece_index = math.floor((stream_pos + file.size - 1) / self._torrent.piece_size)
        return list(range(first_piece_index, last_piece_index + 1))

    def get_files_for_byte_range(self, first_byte_index, last_byte_index):
        """
        Return list of files that have at least one byte at `first_byte_index`,
        `last_byte_index` or between those two in the stream of concatenated
        files
        """
        assert first_byte_index <= last_byte_index, (first_byte_index, last_byte_index)
        pos = 0
        files = []
        for file in self._torrent.files:
            file_first_byte_index = pos
            file_last_byte_index = pos + file.size - 1
            if (
                    # Is first byte of file inside of range?
                    first_byte_index <= file_first_byte_index <= last_byte_index or
                    # Is last byte of file inside of range?
                    first_byte_index <= file_last_byte_index <= last_byte_index or
                    # Are all bytes of file inside of range?
                    (first_byte_index >= file_first_byte_index and last_byte_index <= file_last_byte_index)
            ):
                files.append(file)
            pos += file.size
        return files

    def get_absolute_piece_indexes(self, file, relative_piece_indexes):
        """
        Return list of sanitized absolute piece indexes

        :param file: :class:`~torf.File` object
        :param relative_piece_indexes: Sequence of piece indexes within `file`;
            negative values address pieces at the end of `file`, e.g. [0, 12,
            -1, -2]

        Example:

        >>> # Assume `file` starts in the 50th piece in the stream of
        >>> # concatenated files and is 100 pieces long. `1000` and `-1000` are
        >>> # ignored because they are out of bounds.
        >>> tfs.validate_piece_indexes(file, (0, 1, 70, 75, 1000, -1000, -3, -2, -1))
        [50, 51, 120, 125, 147, 148, 149]
        """
        file_piece_indexes = self.get_piece_indexes_for_file(file)
        pi_abs_min = file_piece_indexes[0]
        pi_abs_max = file_piece_indexes[-1]
        pi_rel_min = 0
        pi_rel_max = pi_abs_max - pi_abs_min

        validated_piece_indexes = set()
        for pi_rel in relative_piece_indexes:
            pi_rel = int(pi_rel)

            # Convert negative to absolute index
            if pi_rel < 0:
                pi_rel = pi_rel_max - abs(pi_rel) + 1

            # Ensure relative piece_index is within bounds
            pi_rel = max(pi_rel_min, min(pi_rel_max, pi_rel))

            # Convert to absolute piece_index
            pi_abs = pi_abs_min + pi_rel
            validated_piece_indexes.add(pi_abs)

        return sorted(validated_piece_indexes)

    def get_piece(self, piece_index):
        """
        Return piece at `piece_index` or `None` if required files don't exist

        :param piece_index: Piece index within stream of concatenated files

        Errors from reading files existing (e.g. :class:`PermissionError`) are
        raised as :class:`~.errors.ContentError`.
        """
        piece_size = self._torrent.piece_size
        torrent_size = sum(f.size for f in self._torrent.files)

        min_piece_index = 0
        max_piece_index = math.floor((torrent_size - 1) / piece_size)
        if not min_piece_index <= piece_index <= max_piece_index:
            raise ValueError(f'piece_index must be in range {min_piece_index} - {max_piece_index}: {piece_index}')

        # Find out which files we need to read from
        first_byte_index_of_piece = piece_index * piece_size
        last_byte_index_of_piece = min(
            first_byte_index_of_piece + piece_size - 1,
            torrent_size - 1,
        )
        relevant_files = self.get_files_for_byte_range(first_byte_index_of_piece, last_byte_index_of_piece)

        # Find out where to start reading in the first relevant file
        if len(relevant_files) == 1:
            # Our piece belongs to a single file
            file_pos = self.get_file_position(relevant_files[0])
            seek_to = first_byte_index_of_piece - file_pos
        else:
            # Our piece is spread over multiple files
            file = self.get_file_at_position(first_byte_index_of_piece)
            file_pos = self.get_file_position(file)
            seek_to = file.size - ((file_pos + file.size) % piece_size)

        # Read piece data from `relevant_files`
        bytes_to_read = piece_size
        piece = bytearray()
        for f in relevant_files:
            fh = self._get_open_file(f)

            # We can't read the piece if any `relevant_file` is missing
            if fh is None:
                return None

            # It's theoretically possible that a differntly sized file can
            # produce the correct piece, but we assume the file is not what
            # we're looking for if it doesn't have the expected size.
            elif f.size != self._get_file_size_from_fs(f):
                return None

            try:
                fh.seek(seek_to)
                seek_to = 0

                content = fh.read(bytes_to_read)
                bytes_to_read -= len(content)
                piece.extend(content)
            except OSError as e:
                msg = e.strerror if e.strerror else e
                raise errors.ContentError(f'Failed to read from {f}: {msg}')

        # Ensure expected `piece` length
        if last_byte_index_of_piece == torrent_size - 1:
            exp_piece_size = torrent_size % piece_size
            if exp_piece_size == 0:
                exp_piece_size = piece_size
        else:
            exp_piece_size = piece_size
        assert len(piece) == exp_piece_size, (len(piece), exp_piece_size)
        return bytes(piece)

    def _get_file_size_from_fs(self, file):
        filepath = os.path.join(self._location, file)
        if os.path.exists(filepath):
            try:
                return os.path.getsize(filepath)
            except OSError:
                # Return None instead of raising an exception
                pass

    def _get_open_file(self, file):
        filepath = os.path.join(self._location, file)
        if filepath not in self._open_files:
            if not os.path.exists(filepath):
                return None
            else:
                try:
                    self._open_files[filepath] = open(filepath, 'rb')
                except OSError as e:
                    msg = e.strerror if e.strerror else e
                    raise errors.ContentError(f'Failed to open {filepath}: {msg}')
        return self._open_files.get(filepath, None)

    def get_piece_hash(self, piece_index):
        """
        Read piece at `piece_index` from file(s) and return its SHA1 hash

        :raise ContentError: if a file exists but cannot be read

        :return: :class:`bytes`
        """
        piece = self.get_piece(piece_index)
        if piece is not None:
            return hashlib.sha1(piece).digest()

    @property
    def torrent_piece_hashes(self):
        """Sequence of piece hashes stored in torrent"""
        pieces = self._torrent.metainfo['info']['pieces']
        piece_hashes = tuple(
            pieces[i : i + 20]
            for i in range(0, len(pieces), 20)
        )
        assert pieces[:20] == piece_hashes[0], (pieces[:20], piece_hashes[0])
        assert pieces[-20:] == piece_hashes[-1], (pieces[-20:], piece_hashes[-1])
        return piece_hashes

    def verify_piece(self, piece_index):
        """
        Generate SHA1 hash for piece at `piece_index` and compare to the expected
        hash in the torrent

        :raise ContentError: if a file exists but cannot be read

        :return: `True` if both hash values are equal, `False` otherwise
        """
        try:
            stored_piece_hash = self.torrent_piece_hashes[piece_index]
        except IndexError:
            max_piece_index = len(self.torrent_piece_hashes) - 1
            raise ValueError(f'piece_index must be in range 0 - {max_piece_index}: {piece_index}')

        generated_piece_hash = self.get_piece_hash(piece_index)
        if generated_piece_hash is not None:
            return stored_piece_hash == generated_piece_hash



"""
Check if scene release was altered
"""

import asyncio
import collections
import difflib
import os
import re

from ... import constants, errors, utils
from ..types import ReleaseType, SceneCheckResult
from . import common, find, srrdb

import logging  # isort:skip
_log = logging.getLogger(__name__)

_srrdb = srrdb.SrrdbApi()

_abbreviated_scene_filename_regexs = (
    # Match names with group in front
    re.compile(r'^[a-z0-9]+-[a-z0-9_\.-]+?(?!-[a-z]{2,})\.(?:mkv|avi)$'),
    # Match "ttl.720p-group.mkv"
    re.compile(r'^[a-z0-9]+[\.-]\d{3,4}p-[a-z]{2,}\.(?:mkv|avi)$'),
    # Match "GR0UP1080pTTL.mkv"
    re.compile(r'^[a-zA-Z0-9]+\d{3,4}p[a-zA-Z]+\.(?:mkv|avi)$'),
)


def assert_not_abbreviated_filename(filepath):
    """
    Raise :class:`~.errors.SceneError` if `filepath` points to an abbreviated
    scene release file name like ``abd-mother.mkv``
    """
    filename = os.path.basename(filepath)
    for regex in _abbreviated_scene_filename_regexs:
        if regex.search(filename):
            raise errors.SceneAbbreviatedFilenameError(filename)


def is_abbreviated_filename(filepath):
    """
    Whether `filepath` points to an abbreviated scene release file name like
    ``abd-mother.mkv``
    """
    try:
        assert_not_abbreviated_filename(filepath)
    except errors.SceneAbbreviatedFilenameError:
        return True
    else:
        return False


@utils.asyncmemoize
async def is_scene_release(release):
    """
    Check if `release` is a scene release or not

    :param release: Release name, path to release or
        :class:`~.release.ReleaseInfo` instance

    :return: :class:`~.types.SceneCheckResult`
    """
    if isinstance(release, str):
        release_info = utils.release.ReleaseInfo(release)
    else:
        release_info = release

    # NOTE: Simply searching for the group does not work because some scene
    #       groups make non-scene releases and vice versa.
    #       Examples:
    #         - Prospect.2018.720p.BluRay.DD5.1.x264-LoRD
    #         - How.The.Grinch.Stole.Christmas.2000.720p.BluRay.DTS.x264-EbP

    results = await find.search(release)
    if results:
        # Do we have enough information to pinpoint a single release?
        needed_keys = common.get_needed_keys(release_info)
        if not needed_keys:
            # If we don't know how to identify a release uniquely, we are not in
            # a position to make any claims.
            return SceneCheckResult.unknown
        elif all(release_info[k] for k in needed_keys):
            return SceneCheckResult.true
        else:
            return SceneCheckResult.unknown

    # If this is a file like "abd-mother.mkv", it's possibly a scene release,
    # but we can't be sure. If it is placed in a properly named parent
    # directory, ReleaseInfo() should've picked that up above.
    if is_abbreviated_filename(release_info.path):
        return SceneCheckResult.unknown

    return SceneCheckResult.false


async def release_files(release_name):
    """
    Map release file names to file information

    This function uses :func:`~.find.search` for searching and
    :class:`~.srrdb.SrrdbApi` to get the file information.

    :param str release_name: Exact name of the release
    """
    files = await _srrdb.release_files(release_name)
    if files:
        return files
    else:
        _log.debug('No such release: %r', release_name)
        files = {}

    # If scene released "Foo.S01E0{1,2,3,...}.720p.BluRay-BAR" and we're
    # searching for "Foo.S01.720p.BluRay-BAR", we might not get any results. But
    # we can get release names of individual episodes by searching for the
    # season pack, and then we can call release_files() for each episode.
    release_info = utils.release.ReleaseInfo(release_name)
    if release_info['type'] is ReleaseType.season:
        results = await find.search(release_info)
        if results:
            files = await asyncio.gather(
                *(_srrdb.release_files(result) for result in results)
            )
            # Join sequence of dictionaries into single dictionary
            files = {fname:f for files_ in files for fname,f in files_.items()}
            _log.debug('Season pack from multiple episode releases: %r', files)

    # If scene released season pack and we're searching for a single episode, we
    # might not get any results. Search for the season pack to get all files.
    elif release_info['type'] is ReleaseType.episode:
        # Remove single episodes from seasons
        release_info['episodes'].remove_specific_episodes()
        results = await find.search(release_info)
        if len(results) == 1:
            _log.debug('Getting files from single result: %r', results[0])
            files = await _srrdb.release_files(results[0])

    # Go through all files and find the exact release name we're looking for.
    # Don't do this exclusively for episodes because not all multi-file releases
    # are a list of episodes (e.g. extras may not contain any "Exx").
    for file_name, file_info in files.items():
        if utils.fs.strip_extension(release_name) == utils.fs.strip_extension(file_name):
            files = {file_name: file_info}
            _log.debug('Single file from season pack release: %r', files)
            break

    return files


async def verify_release_name(content_path, release_name):
    """
    Check if release was renamed

    :param content_path: Path to release file or directory (does not have to
        exist)
    :param release_name: Known exact release name, e.g. from :func:`search`
        results

    :raise SceneRenamedError: if release was renamed
    :raise SceneError: if release name is not a scene release
    """
    _log.debug('Verifying release name: %r =? %r', content_path, release_name)

    if not await is_scene_release(release_name):
        raise errors.SceneError(f'Not a scene release: {release_name}')

    files = await release_files(release_name)
    content_path = content_path.strip(os.sep)

    # Figure out which file is the actual payload. Note that the release may not
    # contain any files, e.g. Wrecked.2011.DiRFiX.LIMITED.FRENCH.720p.BluRay.X264-LOST.
    main_release_file = None
    if files:
        content_file_extension = utils.fs.file_extension(content_path)
        if content_file_extension:
            # `content_path` points to a file, not a directory
            content_file = utils.fs.basename(content_path)
            # We don't know if `content_file` was renamed, so we find the match
            # in the actual file names
            filename_matches = difflib.get_close_matches(content_file, files)
            if filename_matches:
                main_release_file = filename_matches[0]

        if not main_release_file:
            # Default to the largest file if the release has files
            main_release_file = sorted(
                (info for info in files.values()),
                key=lambda info: info['size'],
            )[0]['file_name']

    # No files in this release, default to `release_name`
    if not main_release_file:
        main_release_file = release_name
    _log.debug('Main release file: %r', main_release_file)

    # Generate list of paths that are valid for this release
    acceptable_paths = {release_name}

    # Properly named directory that contains the released file. This covers
    # abbreviated files and all other files.
    for file in files:
        acceptable_paths.add(os.path.join(release_name, file))

    # Any non-abbreviated files may exist outside of a properly named parent
    # directory
    for file in files:
        if not is_abbreviated_filename(file):
            acceptable_paths.add(file)

    # If `release_name` is an episode, it may be inside a season pack parent
    # directory. This only matters if we're dealing with an abbreviated file
    # name; normal file names are independent of their parent directory name.
    if is_abbreviated_filename(content_path):
        season_pack_name = re.sub(
            (
                rf'((?i:{utils.release.DELIM}|S\d+))'
                r'(?i:E\d+)+'
                rf'((?i:{utils.release.DELIM}|$))'
            ),
            r'\1\2',
            release_name,
        )
        for file in files:
            acceptable_paths.add(f'{season_pack_name}/{file}')

    # Standalone file is also ok if it is named `release_name` with the same
    # file extension as the main file
    main_release_file_extension = utils.fs.file_extension(main_release_file)
    acceptable_paths.add('.'.join((release_name, main_release_file_extension)))

    # Release is correctly named if `content_path` ends with any acceptable path
    for path in (p.strip(os.sep) for p in acceptable_paths):
        if re.search(rf'(?:^|{re.escape(os.sep)}){re.escape(path)}$', content_path):
            return

    # All attempts to match `content_path` against `release_name` have failed.
    # Produce a useful error message.
    if is_abbreviated_filename(content_path):
        # Abbreviated files should never be handled without a parent
        original_name = os.path.join(release_name, main_release_file)
    elif utils.fs.file_extension(content_path):
        # Assume `content_path` refers to a file, not a directory
        # NOTE: We can't use os.path.isdir(), `content_path` may not exist.
        original_name = main_release_file
    else:
        # Assume `content_path` refers to directory
        original_name = release_name

    # Use the same number of parent directories for original/existing path. If
    # original_name contains the parent directory, we also want the parent
    # directory in existing_name.
    original_name_parts_count = original_name.count(os.sep)
    content_path_parts = content_path.split(os.sep)
    existing_name = os.sep.join(content_path_parts[-original_name_parts_count - 1:])

    raise errors.SceneRenamedError(
        original_name=original_name,
        existing_name=existing_name,
    )


async def verify_release_files(content_path, release_name):
    """
    Check if files released by scene have the correct size

    :param content_path: Path to release file or directory
    :param release_name: Known exact release name, e.g. from :func:`search`
        results

    The return value is a sequence of :class:`~.errors.SceneError` exceptions:

        * :class:`~.errors.SceneFileSizeError` if a file has the wrong size
        * :class:`~.errors.SceneMissingInfoError` if information about a file
          cannot be found
        * :class:`~.errors.SceneError` if `release_name` is not a scene release
    """
    exceptions = []

    if not await is_scene_release(release_name):
        exceptions.append(errors.SceneError(f'Not a scene release: {release_name}'))
    if exceptions:
        return tuple(exceptions)

    fileinfos = await release_files(release_name)

    def get_release_filesize(filename):
        return fileinfos.get(filename, {}).get('size', None)

    # Map file paths to expected file sizes
    if os.path.isdir(content_path):
        exp_filesizes = {filepath: get_release_filesize(utils.fs.basename(filepath))
                         for filepath in utils.fs.file_list(content_path)}
    else:
        if len(fileinfos) == 1:
            filename = tuple(fileinfos)[0]
            exp_filesize = get_release_filesize(filename)
            exp_filesizes = {
                # Title.2015.720p.BluRay.x264-FOO.mkv
                content_path: exp_filesize,
                # Title.2015.720p.BluRay.x264-FOO/foo-title.mkv
                os.path.join(utils.fs.strip_extension(content_path), filename): exp_filesize,
            }
        else:
            filename = utils.fs.basename(content_path)
            exp_filesizes = {content_path: get_release_filesize(filename)}

    # Compare expected file sizes to actual file sizes
    _log.debug('File sizes: %r', exp_filesizes)
    for filepath, exp_size in exp_filesizes.items():
        filename = utils.fs.basename(filepath)
        actual_size = utils.fs.file_size(filepath)
        _log.debug('Checking file size: %s: %r ?= %r', filename, actual_size, exp_size)
        if exp_size is None:
            _log.debug('No info: %s', filename)
            exceptions.append(errors.SceneMissingInfoError(filename))
        elif actual_size is not None:
            if actual_size != exp_size:
                _log.debug('Wrong size: %s', filename)
                exceptions.append(
                    errors.SceneFileSizeError(
                        filename=filename,
                        original_size=exp_size,
                        existing_size=actual_size,
                    )
                )
            else:
                _log.debug('Correct size: %s', filepath)
        else:
            _log.debug('No such file: %s', filepath)

    return tuple(e for e in exceptions if e)


async def verify_release(content_path, release_name=None):
    """
    Find matching scene releases and apply :func:`verify_release_name` and
    :func:`verify_release_files`

    :param content_path: Path to release file or directory
    :param release_name: Known exact release name or `None` to :func:`search`
        for `content_path`

    :return: :class:`~.types.SceneCheckResult` enum from
        :func:`is_scene_release` and sequence of :class:`~.errors.SceneError`
        exceptions from :func:`verify_release_name` and
        :func:`verify_release_files`
    """
    if release_name:
        return await _verify_release(content_path, release_name)

    # Find possible `release_name` values. For season packs that were released
    # as single episodes, this will get us a sequence of episode release names.
    existing_release_names = await find.search(content_path)
    if not existing_release_names:
        return SceneCheckResult.false, ()

    # Maybe `content_path` was released by scene as it is (as file or directory)
    for existing_release_name in existing_release_names:
        is_scene_release, exceptions = await _verify_release(content_path, existing_release_name)
        if is_scene_release and not exceptions:
            return SceneCheckResult.true, ()

    # Maybe `content_path` is a directory (season pack) and scene released
    # single files (episodes).
    return await _verify_release_per_file(content_path)


async def _verify_release_per_file(content_path):
    _log.debug('Verifying each file beneath %r', content_path)
    is_scene_releases = []
    combined_exceptions = collections.defaultdict(lambda: [])
    filepaths = utils.fs.file_list(content_path, extensions=constants.VIDEO_FILE_EXTENSIONS)
    for filepath in filepaths:
        existing_release_names = await find.search(filepath)
        _log.debug('Search results for %r: %r', filepath, existing_release_names)

        # If there are no search results, default to "not a scene release"
        is_scene_release = SceneCheckResult.false

        # Match each existing_release_name against filepath
        for existing_release_name in existing_release_names:
            is_scene_release, exceptions = await _verify_release(filepath, existing_release_name)
            _log.debug('Verified %r against %r: %r, %r',
                       filepath, existing_release_name, is_scene_release, exceptions)
            if is_scene_release and not exceptions:
                # Match found, don't check other existing_release_names
                break
            elif is_scene_release:
                # Remember exceptions per file (makes debugging easier)
                combined_exceptions[filepath].extend(exceptions)

        # Remember the SceneCheckResult when the for loop ended. True if we
        # found a scene release at any point, other it's the value of the last
        # existing_release_name.
        is_scene_releases.append(is_scene_release)

    # Collapse `is_scene_releases` into a single value
    if is_scene_releases and all(isr is SceneCheckResult.true for isr in is_scene_releases):
        _log.debug('All files are scene releases')
        is_scene_release = SceneCheckResult.true
    elif is_scene_releases and all(isr is SceneCheckResult.false for isr in is_scene_releases):
        _log.debug('All files are non-scene releases')
        is_scene_release = SceneCheckResult.false
    else:
        _log.debug('Uncertain scene status: %r', is_scene_releases)
        is_scene_release = SceneCheckResult.unknown

    return is_scene_release, tuple(exception
                                   for exceptions in combined_exceptions.values()
                                   for exception in exceptions)


async def _verify_release(content_path, release_name):
    _log.debug('Verifying %r against release: %r', content_path, release_name)

    # Stop other checks if this is not a scene release
    is_scene = await is_scene_release(release_name)
    if not is_scene:
        return SceneCheckResult.false, ()

    # Combine exceptions from verify_release_name() and verify_release_files()
    exceptions = []

    # verify_release_name() can only produce one exception, so it is raised
    try:
        await verify_release_name(content_path, release_name)
    except errors.SceneError as e:
        exceptions.append(e)

    # verify_release_files() can produce multiple exceptions, so it returns them
    exceptions.extend(await verify_release_files(content_path, release_name))
    return is_scene, tuple(exceptions)



"""
Client API for Deluge
"""

import base64
import os
import re

import deluge_client

from ... import errors, utils
from .. import asynccontextmanager
from .base import ClientApiBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class DelugeClientApi(ClientApiBase):
    """
    Deluge API

    References:
        https://github.com/JohnDoee/deluge-client#usage
        https://dev.deluge-torrent.org/wiki/UserGuide/ThinClient
    """

    name = 'deluge'
    label = 'Deluge'

    default_config = {
        'host': '127.0.0.1',
        'port': utils.types.Integer(58846, min=0, max=65535),
        'username': '',
        'password': '',
        'check_after_add': utils.configfiles.config_value(
            value=utils.types.Bool('no'),
            description=f'Whether to tell {label} to verify pieces after adding the torrent.',
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._client = deluge_client.DelugeRPCClient(
            host=self.config['host'],
            port=self.config['port'],
            username=self.config['username'],
            password=self.config['password'],
        )

    async def _call(self, method_name, *args, **kwargs):
        method = getattr(self._client, method_name)
        try:
            return method(*args, **kwargs)
        except deluge_client.client.DelugeClientException as e:
            raise errors.RequestError(e)

    async def _login(self):
        await self._call('connect')

    async def _logout(self):
        await self._call('disconnect')

    @asynccontextmanager
    async def _session(self):
        await self._login()
        try:
            yield
        finally:
            await self._logout()

    async def add_torrent(self, torrent_path, download_path=None):
        try:
            torrent_filedump = str(
                base64.b64encode(self.read_torrent_file(torrent_path)),
                encoding='ascii',
            )
        except errors.TorrentError as e:
            raise errors.RequestError(e)

        options = {
            # Assume that all files are present for this torrent
            'seed_mode': not bool(self.config['check_after_add']),
        }
        if download_path:
            options['download_location'] = str(os.path.abspath(download_path))

        async with self._session():
            try:
                infohash_bytes = await self._call(
                    'core.add_torrent_file',
                    filename=utils.fs.basename(torrent_path),
                    filedump=torrent_filedump,
                    options=options,
                )
            except errors.RequestError as e:
                infohash_regex = re.compile(r'Torrent already in session \(([0-9a-zA-Z]+)\)')
                match = infohash_regex.search(str(e))
                if match:
                    infohash = match.group(1)
                else:
                    raise
            else:
                infohash = infohash_bytes.decode('ascii')

        return infohash



import asyncio
import http
import os
import urllib
import xmlrpc

from .... import errors, utils
from ..base import ClientApiBase
from . import _scgi

import logging  # isort:skip
_log = logging.getLogger(__name__)


class RtorrentClientApi(ClientApiBase):
    """
    rtorrent API

    References:
        https://github.com/rakshasa/rtorrent/wiki/rTorrent-0.9-Comprehensive-Command-list-(WIP)
        https://docs.python.org/3/library/xmlrpc.client.html
        https://github.com/rakshasa/rtorrent/wiki/RPC-Setup-XMLRPC
    """

    name = 'rtorrent'
    label = 'rTorrent'

    default_config = {
        'url': utils.configfiles.config_value(
            value='http://localhost:5000/',
            description=(
                'Formats:\n'
                '  [https|http]://<username>:<password>@<host>:<port>\n'
                '  [https|http]://<host>:<port>\n'
                '  scgi://<host>:<port>\n'
                '  scgi:///path/to/rtorrent/rpc.socket'
            ),
        ),
        'check_after_add': utils.configfiles.config_value(
            value=utils.types.Bool('no'),
            description=f'Whether to tell {label} to verify pieces after adding the torrent.',
        ),
    }

    add_torrent_timeout = 10
    add_torrent_check_delay = 0.5

    @utils.cached_property
    def _proxy(self):
        if not self.config['url']:
            raise errors.RequestError('No URL provided')

        parts = urllib.parse.urlsplit(self.config['url'])
        if parts.scheme == 'scgi':
            if parts.netloc and ':' in parts.netloc:
                # If there is a port in netloc, it's a host.
                kwargs = {
                    'uri': 'http://' + parts.netloc,
                    'transport': _scgi.ScgiTransport(),
                }
            else:
                # If there is no port in the host, it's the first segment of a
                # relative path. For absolute paths, netloc is empty.
                socket_path = parts.netloc + parts.path
                kwargs = {
                    'uri': 'http://1',
                    'transport': _scgi.ScgiTransport(socket_path=socket_path),
                }
        else:
            kwargs = {
                'uri': self.config['url'],
            }
        return xmlrpc.client.ServerProxy(**kwargs)

    def _request(self, method, *args):
        attr = self._proxy
        for name in method.split('.'):
            attr = getattr(attr, name)

        try:
            return attr(*args)
        except http.client.HTTPException as e:
            raise errors.RequestError(e)
        except xmlrpc.client.ProtocolError as e:
            raise errors.RequestError(e.errmsg)
        except OSError as e:
            msg = e.strerror if e.strerror else str(e)
            raise errors.RequestError(msg)

    def _get_torrent_hashes(self):
        return tuple(infohash.lower() for infohash in self._request('download_list'))

    async def _get_load_command(self):
        available_commands = self._request('system.listMethods')
        wanted_commands = (
            'load.raw_start_verbose',
            'load.raw_start',
        )
        for cmd in wanted_commands:
            if cmd in available_commands:
                return cmd
        raise RuntimeError('Failed to find load command')

    async def add_torrent(self, torrent_path, download_path=None):
        _log.info('Adding to rtorrent: %r', torrent_path)
        torrent_path = str(os.path.abspath(torrent_path))
        if download_path:
            download_path = str(os.path.abspath(download_path))

        try:
            args = ['', self._get_torrent_data(torrent_path, download_path)]
        except errors.TorrentError as e:
            raise errors.RequestError(e)

        if download_path:
            args.append('d.directory.set="'
                        + download_path.replace('"', r'\"')
                        + '"')

        self._request(await self._get_load_command(), *args)
        try:
            infohash = self.read_torrent(torrent_path).infohash
        except errors.TorrentError as e:
            raise errors.RequestError(e)

        aioloop = asyncio.get_event_loop()
        start_time = aioloop.time()
        while aioloop.time() - start_time < self.add_torrent_timeout:
            infohashes = self._get_torrent_hashes()
            if infohash in infohashes:
                return infohash
            else:
                await asyncio.sleep(self.add_torrent_check_delay)
        raise errors.RequestError('Unknown error')

    def _get_torrent_data(self, torrent_path, download_path=None):
        if download_path and not self.config['check_after_add']:
            try:
                torrent = self._get_torrent_data_with_resume_fields(torrent_path, download_path)
            except errors.TorrentError as e:
                _log.debug('Failed to generate libtorrent_resume fields: %r', e)
                torrent = self.read_torrent(torrent_path)
        else:
            torrent = self.read_torrent(torrent_path)
        return torrent.dump()

    def _get_torrent_data_with_resume_fields(self, torrent_path, download_path):
        _log.debug('Adding libtorrent_resume fields')
        torrent = self.read_torrent(torrent_path)
        files = []
        with utils.torrent.TorrentFileStream(torrent, download_path) as filestream:
            for file in torrent.files:
                filepath = os.path.join(download_path, file)
                files.append({
                    'mtime': self._get_mtime(filepath),
                    # Number of chunks in this file
                    'completed': len(filestream.get_piece_indexes_for_file(file)),
                    # rtorrent_fast_resume.pl sets priority to 0 while autotorrent
                    # sets it to 1. Not sure what's better or if this matters at
                    # all.
                    'priority': 0,
                })

        piece_count = len(torrent.metainfo['info']['pieces']) // 20
        torrent.metainfo['libtorrent_resume'] = {
            'files': files,
            'bitfield': piece_count,
        }
        return torrent

    def _get_mtime(self, path):
        try:
            return int(os.stat(path).st_mtime)
        except OSError as e:
            msg = e.strerror if e.strerror else str(e)
            raise errors.TorrentError(msg)



"""
Translation between country names and various codes
"""

import functools
import logging


@functools.lru_cache(maxsize=None)
def _get_country_converter():
    import country_converter
    country_converter_logger = logging.getLogger('country_converter')
    country_converter_logger.setLevel(logging.CRITICAL)
    return country_converter.CountryConverter()


def _flatten_sequence(seq):
    return tuple(
        item
        for subseq in seq
        for item in subseq
    )


def _convert(countries, to):
    values = _flatten_sequence(
        _get_country_converter().convert(
            countries,
            to=to,
            enforce_list=True,
            not_found=None,  # Return original input
        )
    )
    if isinstance(countries, str) and len(values) == 1:
        return values[0]
    else:
        return tuple(values)


def country_name(country):
    """
    Convert vague country name to consistent name

    :param country: Country name or code, e.g. "Russian Federation" (official
        name), "USA" (common abbreviation), "Korea" (short name), "fr"
        (2-character country code), etc.
    :type country: str or sequence

    :return: Country name or :class:`tuple` of country names, depending on
        `country` argument, or `country` if it can't be associated with any
        country
    """
    return _convert(country, 'name_short')


def country_code_2letter(country):
    """
    Convert vague country names to ISO 3166-1 alpha-2 codes

    :param country: Country name or code, e.g. "Russian Federation" (official
        name), "USA" (common abbreviation), "Korea" (short name), "fr"
        (2-character country code), etc.
    :type country: str or sequence

    :return: Country code or :class:`tuple` of country codes, depending on
        `country` argument
    """
    return _convert(country, 'ISO2')


def country_tld(country):
    """
    Convert vague country names to top level domains

    This function behaves exactly like :func:`country_codes_2letter` except
    that:

        * Values are lower case.
        * For the United Kingdom, the value is "uk", not "gb".
    """
    def convert(tld):
        return ('uk' if tld == 'GB' else tld).lower()

    tlds = _convert(country, 'ISO2')
    if isinstance(tlds, str):
        return convert(tlds)
    else:
        return tuple(convert(tld) for tld in tlds)



"""
Search for scene release
"""

import collections
import re

from ... import constants, errors
from .. import LazyModule, asyncmemoize, fs, release
from . import common, verify

import logging  # isort:skip
_log = logging.getLogger(__name__)

natsort = LazyModule(module='natsort', namespace=globals())


@asyncmemoize
async def search(query, dbs=('predb', 'srrdb'), only_existing_releases=None):
    """
    Search scene databases

    Try to get search results from multiple :class:`~.base.SceneDbApiBase`
    instances and return the first response.

    Failed requests are ignored unless all requests fail, in which case they are
    combined into a single :class:`~.errors.RequestError`.

    If there are no results and `query` is a directory path that looks like a
    season pack, perform one search per video file in that directory or any
    subdirectory. This is necessary to find mixed season packs.

    :param query: :class:`SceneQuery` object or :class:`str` to pass to
        :meth:`SceneQuery.from_string` or :class:`~.collections.abc.Mapping` to
        pass to :meth:`SceneQuery.from_release`
    :param dbs: Sequence of :attr:`~.base.SceneDbApiBase.name` values
    :param only_existing_releases: See :meth:`.SceneQuery.search`

    :return: Sequence of release names (:class:`str`)

    :raise RequestError: if all search requests fail
    """
    path = None
    if isinstance(query, str):
        path = query
        query = SceneQuery.from_string(query)
    elif isinstance(query, collections.abc.Mapping):
        query = SceneQuery.from_release(query)

    results = await _multisearch(dbs, query, only_existing_releases)
    if results:
        return results
    elif path:
        combined_results = []
        for episode_query in _generate_episode_queries(path):
            results = await _multisearch(dbs, episode_query, only_existing_releases)
            combined_results.extend(results)
        return combined_results


async def _multisearch(dbs, query, only_existing_releases):
    # Send the same query to multiple DBs
    from . import scenedb

    exceptions = []
    for db_name in dbs:
        db = scenedb(db_name)
        try:
            results = await db.search(query, only_existing_releases=only_existing_releases)
        except errors.RequestError as e:
            _log.debug('Collecting scene search error: %r', e)
            exceptions.append(e)
        else:
            _log.debug('Returning scene search results: %r', results)
            return results

    if exceptions:
        msg = 'All queries failed: ' + ', '.join(str(e) for e in exceptions)
        raise errors.RequestError(msg)
    else:
        return []


def _generate_episode_queries(path):
    info = release.ReleaseInfo(path)
    if info['type'] is release.ReleaseType.season:
        # Create `query` from each `episode_path`
        episode_paths = fs.file_list(path, extensions=constants.VIDEO_FILE_EXTENSIONS)
        for episode_path in episode_paths:
            _log.debug('Creating query for episode: %r', episode_path)
            if not verify.is_abbreviated_filename(episode_path):
                yield SceneQuery.from_string(episode_path)


class SceneQuery:
    """
    Query for scene release databases

    :param keywords: Search keywords
    :param group: Release group name
    :param episodes: :class:`~.release.Episodes`-like mapping
    """

    @classmethod
    def from_string(cls, string):
        """
        Create query from `string`

        :param string: Release name or path to release content

        `string` is passed to :class:`~.release.ReleaseInfo` and the result is
        passed to :meth:`from_release`.

        :raise SceneError: if :class:`~.release.ReleaseInfo` raises
            :class:`~.errors.ContentError`
        """
        try:
            return cls.from_release(release.ReleaseInfo(string, strict=False))
        except errors.ContentError as e:
            raise errors.SceneError(e)

    @classmethod
    def from_release(cls, release):
        """
        Create query from :class:`dict`-like object

        :param release: :class:`~.release.ReleaseName` or
            :class:`~.release.ReleaseInfo` instance or any :class:`dict`-like
            object with the keys ``title``, ``year``, ``episodes``,
            ``resolution``, ``source``, ``video_codec`` and ``group``.
        """
        # Replace H.264/5 with H264/5
        release['video_codec'] = re.sub(r'\.', '', release.get('video_codec', ''))

        # Replace WEB-DL with WEB
        if release.get('source') == 'WEB-DL':
            release['source'] = 'WEB'

        # Group and episodes are handled separately from the other keywords
        needed_keys = [k for k in common.get_needed_keys(release)
                       if k not in ('group', 'episodes')]
        keywords = [release[key] for key in needed_keys
                    if key in release]

        # Add single episode ("SxxEyy"), but not season pack ("Sxx") or anything
        # else (e.g. "SxxEyyEzz")
        seasons = tuple(release.get('episodes', {}).keys())
        episodes = tuple(episode
                         for episodes in release.get('episodes', {}).values()
                         for episode in episodes)
        if len(seasons) == 1 and len(episodes) == 1:
            keywords.append(f'S{int(seasons[0]):02d}E{int(episodes[0]):02d}')

        query = cls(
            *keywords,
            group=release.get('group', ''),
            episodes=release.get('episodes', {}),
        )
        _log.debug('SceneQuery from %r: %r', release, query)
        return query

    def __init__(self, *keywords, group='', episodes={}):
        # Split each keyword at spaces
        kws = (k.strip()
               for kw in keywords
               for k in str(kw).split())

        # Remove season packs (keep "SxxEyy" but not "Sxx") because scene
        # releases are not season packs and we don't find anything if we look
        # for "S05". For season packs, we try to find all episodes of all
        # seasons and extract the relevant episodes in _handle_results().
        def exclude_season_pack(kw):
            if release.Episodes.is_episodes_info(kw):
                episodes = release.Episodes.from_string(kw)
                for season in tuple(episodes):
                    if not episodes[season]:
                        del episodes[season]
                return str(episodes)
            else:
                return kw

        kws = (exclude_season_pack(kw) for kw in kws)

        # Remove empty keywords
        # (I forgot why "-" is removed. Please explain if you know!)
        kws = (kw for kw in kws
               if kw and kw != '-')

        self._keywords = tuple(kws)
        self._group = str(group) if group else None
        self._episodes = episodes

    async def search(self, search_coro_func, only_existing_releases=None):
        """
        Pre-process query and post-process results from `search_coro_func`

        This is a wrapper function that tries to handle season/episode
        information more intuitively.

        :param search_coro_func: Coroutine function with the call signature
            `(keywords, group)` that returns a sequence of release names
        :param bool only_existing_releases: If this is truthy (the default), the
            results contain all episodes for every season pack in
            :attr:`keywords`.  Otherwise, fake season pack release names are
            included in the returned results.
        """
        if only_existing_releases is None:
            only_existing_releases = True
        results = await search_coro_func(self.keywords, group=self.group)
        return self._handle_results(results, only_existing_releases)

    def _handle_results(self, results, only_existing_releases):
        def sorted_and_deduped(results):
            return natsort.natsorted(set(results), key=str.casefold)

        if not self.episodes:
            _log.debug('No episodes info: %r', self.episodes)
            return sorted_and_deduped(results)
        else:
            _log.debug('Looking for existing episodes: %r', self.episodes)

            def get_wanted_episodes(season):
                # Combine episodes from any season with episodes from given
                # season (season being empty string means "any season")
                eps = None
                if '' in self.episodes:
                    eps = self.episodes['']
                if season in self.episodes:
                    eps = (eps or []) + self.episodes[season]
                return eps

            # Translate single episodes into season packs.
            matches = []
            for result in results:
                for result_season, result_eps in release.Episodes.from_string(result).items():
                    wanted_episodes = get_wanted_episodes(result_season)

                    # [] means season pack
                    if wanted_episodes == []:
                        if only_existing_releases:
                            # Add episode from wanted season pack
                            _log.debug('Adding episode from season pack: %r', result)
                            matches.append(result)
                        else:
                            season_pack = self._create_season_pack_name(result)
                            _log.debug('Adding season pack: %r', season_pack)
                            matches.append(season_pack)

                    elif wanted_episodes is not None:
                        for ep in result_eps:
                            if ep in wanted_episodes:
                                matches.append(result)
                                break

            return sorted_and_deduped(matches)

    @staticmethod
    def _create_season_pack_name(release_name):
        # Remove episode(s) from release name to create season pack name
        season_pack = re.sub(r'\b(S\d{2,})*(E\d{2,})*\b', r'\1', release_name)

        # Remove episode title
        release_info = release.ReleaseInfo(release_name)
        if release_info['episode_title']:
            # Escape special characters in each word and
            # join words with "space or period" regex
            episode_title_regex = rf'{release.DELIM}+'.join(
                re.escape(word)
                for word in release_info['episode_title'].split()
            )
            season_pack = re.sub(rf'\b{episode_title_regex}\W', '', season_pack)

        return season_pack

    @property
    def keywords(self):
        """Sequence of search terms"""
        return self._keywords

    @property
    def group(self):
        """Release group name"""
        return self._group

    @property
    def episodes(self):
        """:class:`~.release.Episodes`-like mapping"""
        return self._episodes

    def __repr__(self):
        args = []
        if self.keywords:
            args.append(', '.join((repr(kw) for kw in self.keywords)))
        if self.group:
            args.append(f'group={self.group!r}')
        if self.episodes:
            args.append(f'episodes={self.episodes!r}')
        args_str = ', '.join(args)
        return f'{type(self).__name__}({args_str})'

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self.keywords == other.keywords
                and self.group == other.group
                and self.episodes == other.episodes
            )
        else:
            return NotImplemented
